;
;	Software regulated bench power supply
;
;	Full PID implementation of voltage regulation
;

	list p=16f870
	#include <p16F870.inc>	
	errorlevel -302	
	__CONFIG _PWRTE_OFF & _WDTE_ON & _HS_OSC

#define SRDAT	PORTC,1
#define SRCLK	PORTC,3
#define LEDON	PORTC,7


	cblock	20h
	
butime	:2
func	:1
vsave	:1
xdata	:1
temp	:1
temp1	:1
count	:1
pwidth	:1
lpcnt	:1
lpxnt	:1
offcnt	:1
slowf	:1
hi		:1
xlow	:1
zhi		:1
zlow	:1
thi		:1
tlow	:1
adjhi	:1
adjlo	:1
vhi		:1
vlo		:1
lasthi	:1
lastlo	:1
alo		:1
ahi		:1
amax	:1
pwmhi	:1
pwmlo	:1
vset	:1
xvset	:1
shaft	:1
del0	:1
del1	:1
clean	:1
seqnr	:1
vsegs	:3
asegs	:3

flag	:1
flag1	:1

	endc

#define	NZ		flag,0
#define	OKBUT	flag,1		;ok for next button
#define	BHIT	flag,2		;button is hit
#define	DOWDT	flag,3		;enable watchdog in delay loop
#define	BUTON	flag,4		;button state
#define BHELD	flag,5		;button held down

#define	RINC	flag1,0		;rotor increment
#define	RDEC	flag1,1		;rotor decrement

	org	0
	
	goto	start
	
	org	4
	retfie
	
drvseg	addwf	PCL,f
	dt	90h,0A0h,0B0h,0D0h,0E0h,0F0h

digseg	addwf	PCL,f
	dt	3fh,6h,5bh,4fh,66h,6dh,7dh,7h,7fh,67h

funcs	addwf	PCL,f
	dt	77h,38h,6	;ALI


start	clrf	STATUS
	movlw	0Ch		;set pwm mode
	movwf	CCP1CON
	bsf	STATUS,5
	movlw	b'00011111'		;set for analog inputs
	movwf	TRISA
	movlw	7
	movwf	TRISB
	movlw	0
	movwf	TRISC
	movlw	.255
	movwf	PR2		;set pwm period
	movlw	80h		;right justify 5 analog input channels
	movwf	ADCON1
	bcf	STATUS,5
	movlw	4
	movwf	T2CON
	clrf	pwmlo
	clrf	pwmhi
	clrf	lasthi
	clrf	lastlo
	clrf	alo
	clrf	ahi
	clrf	seqnr
	clrf	xdata
	clrf	vset
	movlw	.200
	movwf	amax
	movf	PORTB,w
	movwf	shaft


cycle	clrwdt
	call	getvlt		;read the voltage
	call	efac2		;get the proportional error
	call	clamp		;limit amount
	call	newpwm		;+/- to power
	call	diffi		;get the differencial error
	call	newpwm		;adjust power
	call	chekov		;check limits
	call	slowp		;do the integral
	call	setpwm		;send to the hardware
	decfsz	lpcnt,f
	goto	cycle
	movlw	20
	movwf	lpcnt
	call	readsi		;read the input
	call	nxtdsp
	call	dispv	
	call	calcv
	call	onoff
	call	getamp		;read amps
	decfsz	lpxnt,f
	goto	cycle
	movlw	.200
	movwf	lpxnt
	call	dispa		;display amps
	goto	cycle


onoff	btfsc	BUTON
	goto	byp
	movf	alo,w
	movwf	xlow
	movf	ahi,w
	movwf	hi
	rrf	hi,f
	rrf	xlow,f
	movf	amax,w
	subwf	xlow,w
	btfsc	STATUS,C
	goto	off		;power off.
byp	call	button
	btfss	BHIT
	return
	bcf	BHIT
off	clrf	hi
	clrf	xlow
	call	dopwm		;power off
wait1	clrwdt
	call	button
	btfss	BHIT
	goto	cont
	bcf	BHIT
	movf	PORTB,w
	movwf	shaft
	return
cont	movlw	50
	call	delay1
	clrf	vsegs
	clrf	vsegs+1
	clrf	vsegs+2
	clrf	asegs
	clrf	asegs+1
	clrf	asegs+2
	bsf	asegs+2,7
	call	nxtdsp
	call	chkset
	goto	wait1


button	btfss	PORTB,0		;read switch
	goto	but		;button down - toggle state
	bcf	BUTON
	bcf	BHELD
	bcf	BHIT
	movlw	2
	movwf	butime+1
	clrf	butime
	decfsz	offcnt,f
	return
	bsf	OKBUT
	return
but	movlw	50
	movwf	offcnt
	decfsz	butime,f
	goto	ok
	decfsz	butime+1,f
	goto	ok
	bsf	BHELD
ok	bsf	BUTON
	btfss	OKBUT
	return
	bcf	OKBUT
	bsf	BHIT		;button hit valid
	return


chkset	btfss	BHELD		;check button state
	return
	bcf	BHELD
	clrf	func
wait2	clrwdt
	call	button
	btfsc	BHELD
	goto	back
	call	showf
	call	setmax
	movf	amax,w
	movwf	xlow
	clrf	hi
	bcf	STATUS,C
	rlf	xlow,f
	rlf	hi,f
	movlw	asegs
	call	dexi
	bsf	asegs,7
	call	nxtdsp
	bcf	BHIT
	movlw	50
	call	delay1
	goto	wait2
back	bcf	BHELD
	return

showf	movf	func,w
	addwf	func,w
	addwf	func,w		;func x 3
	movwf	temp
	call	funcs
	movwf	vsegs
	incf	temp,f
	movf	temp,w
	call	funcs
	movwf	vsegs+1
	incf	temp,f
	movf	temp,w
	call	funcs
	movwf	vsegs+2
	return

; read shaft encoder	

readsi	call	rotor
	btfsc	RINC
	goto	more
	btfsc	RDEC
	goto	less1
	return
less1	movf	vset,f		;set z flag
	btfsc	STATUS,Z
	goto	done
	decf	vset,f
	goto	done
more1	incf	vset,f
	movf	vset,w
	xorlw	.201
	btfss	STATUS,Z
	goto	done
	decf	vset,f
done	return

setmax	call	rotor
	btfsc	RINC
	goto	more2
	btfsc	RDEC
	goto	less1
	return
less2	movf	amax,f
	btfsc	STATUS,Z
	return
	decf	amax,f
	return
more2	incf	amax,f
	movf	amax,w
	xorlw	.251
	btfss	STATUS,Z
	return
	decf	amax,f
	return

	

rotor	bcf	RINC
	bcf	RDEC
	movf	PORTB,w		;get quad bits
	movwf	temp
	movwf	xdata
	xorwf	shaft,w
	andlw	6		;isolate bits
	btfsc	STATUS,Z		;match?
	return			;yes - return z-set
	movf	shaft,w
	rlf	temp,f
	xorwf	temp,f
	btfsc	temp,2
	goto	up
down	decf	xvset,f
	goto	ok1
up	incf	xvset,f
ok1	btfss	xdata,1
	goto	skip
	btfss	xdata,2		;both bits low ?
	goto	skip		;no.
	movf	xvset,w
	xorlw	4
	btfsc	STATUS,Z
	goto	more
	movf	xvset,w
	xorlw	-4
	btfsc	STATUS,Z
	goto	less
	goto	done
less	bsf	RDEC
	goto	done1
more	bsf	RINC
done1	clrf	xvset
skip	movf	xdata,w
	movwf	shaft
	return

	

calcv	movf	vset,w
	movwf	xlow
	movlw	.73
	call	mult
	rrf	hi,f
	rrf	xlow,f
	rrf	hi,f
	rrf	xlow,f
	rrf	hi,f
	rrf	xlow,f
	rrf	hi,f
	rrf	xlow,f
	movf	xlow,w
	movwf	adjlo
	movf	hi,w
	andlw	3
	movwf	adjhi
	return


slowp	decfsz	slowf,f		;get timeing factor
	return
	movlw	.20
	movwf	slowf
	movf	vhi,w
	subwf	adjhi,w
	btfsc	STATUS,Z
	goto	chklo
chk	btfss	STATUS,C
	goto	dn
	incf	pwmlo,f
	btfsc	STATUS,Z
	incf	pwmhi,f
	return
dn	decf	pwmlo,f
	movf	pwmlo,w
	xorlw	h'0ff'
	btfsc	STATUS,Z
	decf	pwmhi,f
	movf	pwmlo,w
	xorlw	h'0ff'
	btfss	STATUS,Z
	return
	clrf	pwmlo
	clrf	pwmhi
	return
chklo	movf	vlo,w
	subwf	adjlo,w
	btfss	STATUS,Z		;exact match?
	goto	chk
	return			;no - action.


diffi	clrf	hi
	movf	vlo,w
	subwf	lastlo,w	;find last reading
	movwf	xlow
	btfss	STATUS,C
	decf	hi,f
	movf	vhi,w
	subwf	lasthi,w
	addwf	hi,f
	movf	vlo,w
	movwf	lastlo
	movf	vhi,w
	movwf	lasthi
	return
	

; get error factor in hi/low

efac1	clrf	hi
	movf	vlo,w
	subwf	adjlo,w
	movwf	xlow
	btfss	STATUS,C
	decf	hi,f
	movf	vhi,w
	subwf	adjhi,w
	addwf	hi,f
	return

efac2	call	efac1
	btfsc	hi,7		;negative value?
	goto	neg
	bcf	STATUS,C
	rrf	hi,f
	rrf	xlow,f
	bcf	STATUS,C
	rrf	hi,f
	rrf	xlow,f
	return
neg	rrf	hi,f
	rrf	xlow,f
	bsf	hi,7
	rrf	hi,f
	rrf	xlow,f
	bsf	hi,7
	return
	

clamp	btfsc	hi,7
	return			;only process positive corrections
	movf	hi,w
	btfss	STATUS,Z
	goto	max
	movf	xlow,w
	sublw	.10
	btfss	STATUS,C
	goto	max
	return
max	movlw	.10
	movwf	xlow
	clrf	hi
	return
	
; adjust new PWM value using hi/low

newpwm	movf	xlow,w
	addwf	pwmlo,f
	btfsc	STATUS,C
	incf	pwmhi,f
	movf	hi,w
	addwf	pwmhi,f
	return
	
; check PWM value for in range 0 to 3ff.

chekov	btfsc	pwmhi,7		;underflow?
	goto	min
	movf	pwmhi,w
	andlw	0FCh		;overflow?
	btfsc	STATUS,Z
	return			;no - in range
	movlw	.3
	movwf	pwmhi
	movlw	h'0FF'
	movwf	pwmlo
	return
min	clrf	pwmhi
	clrf	pwmlo
	return


; sets 10 bit pwm value in CCP1CON<5,4>/CCPR1L<7-0>

setpwm	movf	pwmlo,w
	movwf	xlow
	movf	pwmhi,w
	movwf	hi
dopwm	movf	xlow,w
	movwf	temp
	swapf	temp,f
	movf	temp,w
	andlw	h'30'
	iorlw	h'0C'		;set for pwm mode
	movwf	CCP1CON
	rrf	hi,f
	rrf	xlow,f
	rrf	hi,f
	rrf	xlow,f		;hi 8 bits
	movf	xlow,w
	movwf	CCPR1L
	return



; multiply w by low into low/hi

mult	movwf	temp
	movlw	.8
	movwf	count
	movf	xlow,w
	movwf	temp1
	clrf	xlow
	clrf	hi
	movf	temp,w
	bcf	STATUS,C
loop	rrf	temp1,f
	btfsc	STATUS,C
	addwf	hi,f
	rrf	hi,f
	rrf	xlow,f
	decfsz	count,f
	goto	loop
	return



getvlt	movlw	0
	call	atod
	movf	xlow,w
	movwf	vlo
	movf	hi,w
	movwf	vhi
	return

getamp	movlw	1
	call	atod
	movf	xlow,w
	movwf	alo
	movf	hi,w
	movwf	ahi
	return


	

dispv	movf	vset,w
	movwf	xlow
	clrf	hi
	movlw	vsegs
	call	dexi
	bsf	vsegs+1,7
	return
	
	
dispa	movf	alo,w
	movwf	xlow
	movf	ahi,w
	movwf	hi
	movlw	asegs
	call	dexi
	bsf	asegs,7
	return
	
	
; decimal values up to 1000 converted directly into
; segment values

dexi	movwf	FSR		;place to put data
	bcf	NZ
	movlw	0
	movwf	thi
	movlw	.100
	movwf	tlow
	call	digit
	movlw	0
	movwf	thi
	movlw	.10
	movwf	tlow
	call	digit
	movlw	0
	movwf	thi
	movlw	.1
	movwf	tlow
	call	digit
	return

digit	clrf	count
dig1	call	dsub
	btfsc	hi,7
	goto	dig2
	bsf	NZ
	incf	count,f
	goto	dig1
dig2	call	dadd
	movf	count,w
	sublw	.9
	btfss	STATUS,C
	goto	err
	movf	count,w
	call	digseg		;get segment data
done2	movwf	0		;store.
	incf	FSR,f
	return
err	movlw	h'49'
	goto	done2
	
dsub	movf	tlow,w
	subwf	xlow,f
	btfss	STATUS,C
	decf	hi,f
	movf	thi,w
	subwf	hi,f
	return

dadd	movf	tlow,w
	addwf	xlow,f
	btfsc	STATUS,C
	incf	hi,f
	movf	thi,w
	addwf	hi,f
	return

nxtdsp	incf	seqnr,f
	movf	seqnr,w
	xorlw	6		;past end of list?
	btfss	STATUS,Z
	goto	ok2
	clrf	seqnr
ok2	bcf	LEDON		;disable all displays
	movf	seqnr,w
	addlw	vsegs		;base of segment data
	movwf	FSR
	movf	0,w		;get segment data
	movwf	temp
	movlw	8
	movwf	count
shft	bcf	SRDAT
	btfss	temp,7		;hi order first
	bsf	SRDAT		;reverse sense
	bsf	SRCLK
	bcf	SRCLK		;shift it out
	rlf	temp,f
	decfsz	count,f
	goto	shft
	movf	seqnr,w
	call	drvseg
	movwf	temp
	movf	PORTC,w
	andlw	h'0F'
	iorwf	temp,w
	movwf	PORTC		;select new digit
	return

	
; read atod channel set by 'w'
; result in hi/low

atod	movwf	temp		;get channel (0-4)
	bcf	STATUS,C
	rlf	temp,f
	rlf	temp,f
	rlf	temp,f		;put in place
	movlw	h'81'		;/32 clock, power up
	iorwf	temp,w
	movwf	ADCON0
	movlw	2
	call	delay
	bsf	ADCON0,2
	nop
	nop
chk1	btfsc	ADCON0,2	;check for conversion complete
	goto	chk1
	bsf	STATUS,5
	movf	ADRESL,w
	bcf	STATUS,5
	movwf	xlow
	movf	ADRESH,w
	movwf	hi
	return
	

delay1	bsf	DOWDT
delay	movwf	del1
d1	movwf	del0
d2	btfsc	DOWDT
	clrwdt
	decfsz	del0,f
	goto	d2
	decfsz	del1,f
	goto	d1
	bcf	DOWDT
	return
	
	end
	

